<?php
declare(strict_types = 1);

/**
 * Méthodes de gestion du HTML.
 *
 * @license http://www.gnu.org/licenses/gpl.html
 * @link http://www.igalerie.org/
 */
class HTML
{
	/**
	 * Liste des attributs de balises autorisés.
	 * Retourne la regexp pour la valeur de l'attribut $attr.
	 *
	 * @param string $attr
	 *
	 * @return mixed
	 */
	public static function allowedAttrs(string $attr = '')
	{
		$allowed_attrs =
		[
			'alt'      => '[-_\w\d\s,;:.%]+',
			'cite'     => '[-_\w\d\s,;:.%]+',
			'class'    => '[-_a-z0-9\s]+',
			'datetime' => '[-0-9]+',
			'height'   => '\d{1,6}',
			'href'     => '[-_\w\d@&=+?%~./,;:*#]+',
			'hreflang' => '[-a-z]{2,5}',
			'id'       => '[-_a-z0-9]+',
			'lang'     => '[-a-z]{2,5}',
			'max'      => '[-+0-9e.]+',
			'pubdate'  => '[-0-9]+',
			'rel'      => '[-_a-z0-9\s.#]+',
			'src'      => '[-_a-z0-9./:=\?&;]+',
			'style'    => '[-_a-z0-9\s;:#,%.]+',
			'target'   => '[-_a-z0-9]+',
			'title'    => '[-_\w\d\s@&=+!?%°~./,;:*#()\[\]\|]+',
			'value'    => '[-+0-9e.]+',
			'width'    => '\d{1,6}'
		];

		if ($attr === '')
		{
			return $allowed_attrs;
		}

		return $allowed_attrs[$attr] ?? FALSE;
	}

	/**
	 * Retourne la liste des balises autorisées.
	 *
	 * @return array
	 */
	public static function allowedTags(): array
	{
		// Toutes les balises de formulaires, d'objets (script, ...)
		// et de structure de page (body, ...) ne sont pas autorisées.
		return
		[
			'a', 'abbr', 'address', 'article', 'aside', 'blockquote',
			'caption', 'cite', 'code', 'colgroup', 'del', 'dfn',
			'div', 'dd', 'dl', 'dt', 'em', 'figcaption', 'figure',
			'footer', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'header',
			'hr', 'img', 'ins', 'kbd', 'legend', 'li', 'mark', 'nav',
			'ol', 'p', 'pre', 'progress', 'q', 'rt', 'samp',
			'section', 'small', 'span', 'strong', 'sub', 'sup',
			'table', 'tbody', 'td', 'tfoot', 'th', 'thead', 'time',
			'tr', 'ul', 'var'
		];
	}

	/**
	 * Décode le code HTML autorisé d'une chaîne.
	 *
	 * @param string $str
	 *
	 * @return string
	 */
	public static function decode(string $str): string
	{
		$decode = function(array $str): string
		{
			$str = $str[0];

			// "\" forcément plus que suspect.
			if (strstr($str, '\\'))
			{
				return $str;
			}

			// Suppression des chevrons ouvrant et fermant.
			$s = preg_replace('`^&lt;\s*/?\s*|\s*/?\s*&gt;$`', '', $str);

			// Récupèration du nom de la balise.
			$tag = explode(' ', $s, 2)[0];

			// Est-ce une balise autorisée ?
			if (!in_array($tag, self::allowedTags()))
			{
				return $str;
			}

			// Création d'un tableau contenant les attributs de la balise.
			if (!is_string($s = preg_replace('`^' . $tag . '\s*`', '', $s)))
			{
				return $str;
			}
			$attrs = preg_split('`(?<=&quot;)\s+|(?<=\')\s+`', $s, 9, PREG_SPLIT_NO_EMPTY);

			// Si la balise contient des attributs, on les analyse un à un.
			if ($count = count($attrs))
			{
				// Vérification de chaque attribut.
				for ($i = 0; $i < $count; $i++)
				{
					// Récupération du nom et de la valeur de l'attribut.
					$attr = preg_split('`\s*=\s*`', $attrs[$i], 2);
					list($name, $value) = $attr;

					// Est-ce un attribut autorisé ?
					if (!$attr_regexp = self::allowedAttrs($name))
					{
						return $str;
					}

					// Suppression des guillemets.
					$value = preg_replace('`^(\'(.+)\'|&quot;(.+)&quot;)$`', '$2$3', $value);
					if ($value === $attr[1])
					{
						return $str;
					}

					// Filtrage de l'attribut "href".
					if ($name == 'href' && preg_match('`javascript\s*:`i', $value))
					{
						return $str;
					}

					// Filtrage de l'attribut "src".
					if ($name == 'src')
					{
						$regex = '(?:' . preg_quote(GALLERY_HOST) . ')?'
							. preg_quote(CONF_GALLERY_PATH)
							. '/images/[/\da-z_-]+\.(?:jpe?g|gif|png|webp)';
						if (!preg_match('`^' . $regex . '$`i', $value))
						{
							return $str;
						}
					}

					// Vérification de la présence d'entités HTML autres que &#039; et &amp;
					if (strstr(str_replace(['&#039;', '&amp;'], '', $value), '&'))
					{
						return $str;
					}

					// Vérification de la valeur de l'attribut.
					if (!preg_match('`^' . $attr_regexp . '$`ui', $value))
					{
						return $str;
					}
				}
			}

			return htmlspecialchars_decode($str, ENT_QUOTES);
		};

		return preg_replace_callback('`&lt;(?:(?!&gt;).){1,255}&gt;`', $decode, $str);
	}

	/**
	 * Convertit les URLs d'une chaîne en liens HTML.
	 *
	 * @param string $str
	 * @param int $limit
	 *   Nombre maximum de caractères du lien.
	 *
	 * @return string
	 */
	public static function linkify(string $str, int $limit = 0): string
	{
		$regexp = '`((?<!&quot;)(?<!")' . Utility::URLRegexp() . ')`i';
		$str = preg_replace($regexp, '<a href="$1">$1</a>', $str);

		return $limit
			? preg_replace(
				'`(<a[^>]+>)(.{' . $limit . '}).*(</a>)`ui',
				'\\1\\2...\\3',
				$str)
			: $str;
	}

	/**
	 * Ajoute des entités HTML à la place
	 * des espaces multiples et des tabulations.
	 *
	 * @param string $str
	 *
	 * @return string
	 */
	public static function spaces(string $str): string
	{
		return preg_replace_callback('`&lt;(?:(?!&gt;).){1,255}&gt;`', function($str)
		{
			return preg_replace('`<span class="space_[^<]+</span>`', ' ', $str[0]);
		},
		str_replace(
			['	', '    ', '   ', '  '],
			[
				'<span class="space_tab">&#9;</span>',
				'<span class="space_4">&emsp;</span>',
				'<span class="space_3">&ensp;&nbsp;</span>',
				'<span class="space_2">&ensp;</span>'
			],
			$str
		));
	}

	/**
	 * Applique la fonction htmlspecialchars à une chaîne
	 * ou aux valeurs d'un tableau ou d'un objet.
	 *
	 * @param mixed $s
	 * @param bool $double_encode
	 *
	 * @return mixed
	 */
	public static function specialchars(&$s, bool $double_encode = TRUE)
	{
		if (is_string($s))
		{
			$s = Utility::deleteInvisibleChars($s);
			$s = defined('ENT_DISALLOWED')
				? htmlspecialchars($s, ENT_DISALLOWED | ENT_QUOTES, 'UTF-8', $double_encode)
				: htmlspecialchars($s, ENT_QUOTES, 'UTF-8', $double_encode);
		}
		else if (is_array($s))
		{
			array_walk($s, [__CLASS__, __FUNCTION__]);
		}
		return $s;
	}
}
?>